#include <windows.h>
#include <math.h>
#include <vector>

using namespace std;

#include "audioeffectx.h"

#include "Simple_Apu.h"



#define PLUGIN_NAME				"FTplug"
#define PLUGIN_VENDOR			"v0.1 06.05.19 Shiru"
#define PLUGIN_UID				'ftpg'

#define NUM_INPUTS				0
#define NUM_OUTPUTS				2
#define NUM_PROGRAMS			128

#define MIDI_CHANNELS_ALL		16

#define CHUNK_SIZE				(NUM_PROGRAMS*5*ENVELOPE_TEXT_LEN*2)

#define MAX_NAME_LEN			32

#define SYNTH_CHANNELS			8

#define ENVELOPE_TEXT_LEN		1024
#define ENVELOPE_DATA_LEN		2048
#define ENVELOPE_NUM_RANGE		0x8000

#define NTSC_CLOCK				1789773
#define NTSC_FPS				60
#define PAL_CLOCK				1662607
#define PAL_FPS					50

#define MAX_GAIN				200

enum {
	TVSYS_PAL=0,
	TVSYS_NTSC
};

enum {
	CHTYPE_2A03_PULSE=0,
	CHTYPE_2A03_TRIANGLE,
	CHTYPE_2A03_NOISE
};

enum {
	NUM_PARAMS=0
};



enum {
	mEventTypeNote=0,
	mEventTypePitchBend,
	mEventTypeProgram,
	mEventTypeControl
};



struct MidiQueueStruct
{
	VstInt32 type;
	VstInt32 delta;
	VstInt32 channel;
	VstInt32 note;
	VstInt32 velocity;
	VstInt32 program;
	VstInt32 value;

	float depth;
};



enum {
	PARSE_PREFIX_NONE=0,
	PARSE_PREFIX_NEGATIVE
};



struct EnvelopeDataStruct {
	VstInt32 value[ENVELOPE_DATA_LEN];

	VstInt32 len;
	VstInt32 loop_ptr;

	bool has_loop;
	bool error;
};



struct ProgramStruct {

	char Name[MAX_NAME_LEN];

	VstInt32 ChType;
	VstInt32 MultiMode;

	char StrVol    [ENVELOPE_TEXT_LEN];
	char StrArp    [ENVELOPE_TEXT_LEN];
	char StrPitch  [ENVELOPE_TEXT_LEN];
	char StrHiPitch[ENVELOPE_TEXT_LEN];
	char StrDuty   [ENVELOPE_TEXT_LEN];

	EnvelopeDataStruct EnvVol;
	EnvelopeDataStruct EnvArp;
	EnvelopeDataStruct EnvPitch;
	EnvelopeDataStruct EnvHiPitch;
	EnvelopeDataStruct EnvDuty;
};



struct ChannelStruct {

	VstInt32 midi_ch;
	VstInt32 type;
	VstInt32 key;			//key that has been pressed
	VstInt32 note;			//note that is playing (not necessarily the same as key pressed)
	VstInt32 velocity;
	
	EnvelopeDataStruct *env_vol;
	EnvelopeDataStruct *env_arp;
	EnvelopeDataStruct *env_pitch;
	EnvelopeDataStruct *env_hipitch;
	EnvelopeDataStruct *env_duty;

	VstInt32 ptr_vol;
	VstInt32 ptr_arp;
	VstInt32 ptr_pitch;
	VstInt32 ptr_hipitch;
	VstInt32 ptr_duty;

	VstInt32 arp_offset;
	VstInt32 vol_output;
	VstInt32 duty_output;
	VstInt32 pitch_acc;

	VstInt32 period_prev;

};




#define TAG(s)		(float)(((s)[0])|(((s)[1])<<7)|(((s)[2])<<14)|(((s)[3]<<21)))

#define DATA		TAG("DATA")
#define MPRG		TAG("MPRG")
#define PROG		TAG("PROG")
#define NAME		TAG("NAME")
#define TYPE		TAG("TYPE")
#define MMOD		TAG("MMOD")
#define EVOL		TAG("EVOL")
#define EARP		TAG("EARP")
#define EPTC		TAG("EPTC")
#define EHPT		TAG("EHPT")
#define EDTY		TAG("EDTY")
#define FSYS		TAG("FSYS")
#define GAIN		TAG("GAIN")
#define DONE		TAG("DONE")



//name,chtype,multimode,five sequences for vol,arp,pitch,hipitch,duty
const char* const PresetData[]={
	"Plain Pulse 12.5%","0","0","15","0","0","0","0",
	"Plain Pulse 25.0%","0","0","15","0","0","0","1",
	"Plain Pulse 50.0%","0","0","15","0","0","0","2",
	"Plain Triangle","1","0","15","0","0","0","0",
	"Plain White Noise","2","0","15","0","0","0","0",
	"Plain Periodic Noise","2","0","15","0","0","0","1",
	"Pulse Solo","0","0","15","12 0","0 0 0 0 0 0 0 | 2 2 2 2 -2 -2 -2 -2","0","2",
	"Tri Kick","1","0","15 15 15 15 0","0","| 200","0","0",
	"Noise Snare","2","0","15 14 12 12 10 10 8 8 8 8 0","4 0 0 -4 -4 -4 -4 -1","0","0","0",
	"\r"
};



class FTPLUG:public AudioEffectX
{
public:

	FTPLUG(audioMasterCallback audioMaster);

	~FTPLUG();

	virtual void setParameter(VstInt32 index,float value);
	virtual float getParameter(VstInt32 index);
	virtual void getParameterLabel(VstInt32 index,char *label);
	virtual void getParameterDisplay(VstInt32 index,char *text);
	virtual void getParameterName(VstInt32 index,char *text); 

	virtual VstInt32 getProgram();
	virtual void setProgram(VstInt32 program);
	virtual void getProgramName(char* name);
	virtual void setProgramName(char* name);

	virtual bool getEffectName(char *name) { strcpy(name,PLUGIN_NAME); return true; }
	virtual bool getVendorString(char *text) { strcpy(text,PLUGIN_VENDOR); return true; }
	virtual VstInt32 getVendorVersion() { return 1000; } 

	VstInt32 canDo(char* text);
	VstInt32 getNumMidiInputChannels(void);
	VstInt32 getNumMidiOutputChannels(void);

	virtual VstInt32 getChunk(void** data,bool isPreset=false);
	virtual VstInt32 setChunk(void* data,VstInt32 byteSize,bool isPreset=false);

	VstInt32 processEvents(VstEvents* ev);
	void processReplacing(float**inputs,float **outputs,VstInt32 sampleFrames);

	VstInt32 MidiChannel;
	VstInt32 MidiProgram[MIDI_CHANNELS_ALL];
	VstInt32 MidiProgramPrev[MIDI_CHANNELS_ALL];

	const char* CompileEnvelope(char *str,EnvelopeDataStruct *env,bool hold);
	const char* CompileInstruments(void);

	ProgramStruct Program[NUM_PROGRAMS];

	void SetUpdateGUI(void);
	void UpdateGUI(void);

	float SampleRatePrev;

	VstInt32 SysType;
	VstInt32 SysClock;
	VstInt32 SysFrameRate;
	VstInt32 Polyphony;			//max number of active channels, 1 for mono mode
	VstInt32 OutputGain;		//100 normal volume, <100 reduce, >100 amplify

protected:

	VstInt32 SavePresetChunk(float *chunk);
	void LoadPresetChunk(float *chunk);

	VstInt32 SaveStringChunk(char *str,float *dst);
	VstInt32 LoadStringChunk(char *str,int max_length,float *src);

	float Chunk[CHUNK_SIZE];

	void MidiAddNote(VstInt32 delta,VstInt32 channel,VstInt32 note,VstInt32 velocity);
	void MidiAddPitchBend(VstInt32 delta,VstInt32 channel,float depth);
	void MidiAddProgramChange(VstInt32 delta,VstInt32 channel,VstInt32 program);
	void MidiAddControl(VstInt32 delta,VstInt32 channel,VstInt32 value);

	bool MidiIsAnyKeyDown(void);

	vector<MidiQueueStruct> MidiQueue;

	unsigned char MidiKeyState[128];

	float MidiPitchBend[MIDI_CHANNELS_ALL];
	float MidiPitchBendRange[MIDI_CHANNELS_ALL];

	VstInt32 MidiRPNLSB;
	VstInt32 MidiRPNMSB;
	VstInt32 MidiDataLSB;
	VstInt32 MidiDataMSB;

	ChannelStruct SynthChannel[SYNTH_CHANNELS];

	VstInt32 SynthAllocateVoice(VstInt32 midi_ch,VstInt32 note);
	void SynthChannelChangeNote(VstInt32 chn,VstInt32 midi_ch,VstInt32 note,VstInt32 velocity);
	void SynthChannelReleaseNote(VstInt32 midi_ch,VstInt32 note);
	void SynthChannelUpdate(VstInt32 chn);

	bool UpdateGuiFlag;
	
	Simple_Apu APU[SYNTH_CHANNELS];
};
