#include "FTPlug.h"
#include "GUI.h"



extern void* hInstance;

inline HINSTANCE GetInstance() { return (HINSTANCE)hInstance; }



const char* GUI::GetClassName(void)
{
	sprintf(ClassName,"FTPLUG%08x",(long)this);

	return (const char*)ClassName;
}



GUI::GUI(AudioEffect* effect)
{
	WNDCLASS wc;

	ftplug=(FTPLUG*)effect;

	rect.left  =0;
	rect.top   =0;
	rect.right =GUI_WINDOW_WDT;
	rect.bottom=GUI_WINDOW_HGT;

	wc.style=CS_GLOBALCLASS;
	wc.lpfnWndProc=WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=GetInstance();
	wc.hIcon=0;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=GetSysColorBrush(COLOR_BTNFACE);
	wc.lpszMenuName=0;
	wc.lpszClassName=GetClassName();

	RegisterClass(&wc);
}



GUI::~GUI()
{
	UnregisterClass(GetClassName(),GetInstance());
}
	
	
	
bool GUI::open(void* ptr)
{
	HINSTANCE hinst;
	int x,y,w,h,style;

	AEffEditor::open(ptr);

	InitCommonControls();

	hinst=GetInstance();

	hWndForm=CreateWindow(GetClassName(),"FTPLUG",WS_CHILD|WS_VISIBLE,0,0,(rect.right-rect.left),(rect.bottom-rect.top),HWND(ptr),0,hinst,this);

	if(!hWndForm) return false;

	SetWindowLong(hWndForm,GWLP_USERDATA,(long)this);

	style=WS_CHILD|WS_VISIBLE;

	x=10;
	y=12;
	w=60;
	h=25;

	hWndLabelV=CreateWindow("static","ST_U",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndLabelA=CreateWindow("static","ST_U",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndLabelP=CreateWindow("static","ST_U",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndLabelH=CreateWindow("static","ST_U",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndLabelD=CreateWindow("static","ST_U",style,x,y,w,h,hWndForm,NULL,hinst,NULL);

	SetWindowText(hWndLabelV,"Volume:");
	SetWindowText(hWndLabelA,"Arp:");
	SetWindowText(hWndLabelP,"Pitch:");
	SetWindowText(hWndLabelH,"Hi-Pitch:");
	SetWindowText(hWndLabelD,"Duty:");

	style=WS_CHILD|WS_VISIBLE|ES_LEFT|WS_BORDER|ES_AUTOHSCROLL;

	x=80;
	y=10;
	w=GUI_WINDOW_WDT-40-x;
	h=25;

	hWndEditV=CreateWindow("edit","0",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndEditA=CreateWindow("edit","0",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndEditP=CreateWindow("edit","0",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndEditH=CreateWindow("edit","0",style,x,y,w,h,hWndForm,NULL,hinst,NULL); y+=30;
	hWndEditD=CreateWindow("edit","0",style,x,y,w,h,hWndForm,NULL,hinst,NULL);

	SendMessage(hWndEditV,EM_SETLIMITTEXT,ENVELOPE_TEXT_LEN,0);
	SendMessage(hWndEditA,EM_SETLIMITTEXT,ENVELOPE_TEXT_LEN,0);
	SendMessage(hWndEditP,EM_SETLIMITTEXT,ENVELOPE_TEXT_LEN,0);
	SendMessage(hWndEditH,EM_SETLIMITTEXT,ENVELOPE_TEXT_LEN,0);
	SendMessage(hWndEditD,EM_SETLIMITTEXT,ENVELOPE_TEXT_LEN,0);

	SetWindowLongPtr(hWndEditV,GWLP_USERDATA,0xdeadf00b);	//Reaper hack to prevent hijacking Space button
	SetWindowLongPtr(hWndEditA,GWLP_USERDATA,0xdeadf00b);
	SetWindowLongPtr(hWndEditP,GWLP_USERDATA,0xdeadf00b);
	SetWindowLongPtr(hWndEditH,GWLP_USERDATA,0xdeadf00b);
	SetWindowLongPtr(hWndEditD,GWLP_USERDATA,0xdeadf00b);

	style=WS_VISIBLE|WS_CHILD|BS_DEFPUSHBUTTON|BS_OWNERDRAW|BS_NOTIFY;

	x=80;
	y=10+5*30+10;
	w=60;
	h=24;

	hWndButtonPulse=CreateWindow("button","Pulse",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_PULSE,hinst,NULL);

	x+=w+5;

	hWndButtonTri=CreateWindow("button","Triangle",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_TRIANGLE,hinst,NULL);

	x+=w+5;

	hWndButtonNoise=CreateWindow("button","Noise",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_NOISE,hinst,NULL);

	x+=w+30;

	hWndButtonMultiMode=CreateWindow("button","Multi",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_MULTI,hinst,NULL);

	x+=w+30;

	hWndButtonNTSC=CreateWindow("button","NTSC",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_NTSC,hinst,NULL);

	x+=w+5;
	
	hWndButtonPAL=CreateWindow("button","PAL",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_PAL,hinst,NULL);

	x=GUI_WINDOW_WDT-30;
	y=0;
	w=30;
	h=200;

	hWndTrackGain=CreateWindowEx(0,TRACKBAR_CLASS,"Gain",WS_CHILD|WS_VISIBLE|TBS_VERT|TBS_NOTICKS|TBS_BOTH,x,y,w,h,hWndForm,(HMENU)IDC_TRACKBAR_GAIN,hinst,NULL );

	SendMessage(hWndTrackGain,TBM_SETRANGE,(WPARAM)TRUE,(LPARAM)MAKELONG(0,MAX_GAIN));
    SendMessage(hWndTrackGain,TBM_SETPAGESIZE,0,(LPARAM)4);  
 
	x=10;
	y=200;
	w=70;
	h=25;

	style=WS_CHILD|WS_VISIBLE;

	hWndLabelMIDI=CreateWindow("static","ST_U",style,x,y+4,w,h,hWndForm,(HMENU)IDC_LABEL_MIDI,hinst,NULL);

	x+=w;
	w=25;

	style=WS_VISIBLE|WS_CHILD|BS_DEFPUSHBUTTON|BS_OWNERDRAW|BS_NOTIFY;

	hWndButtonMIDIUp=CreateWindow("button","+",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_MIDI_UP,hinst,NULL);

	x+=w+5;

	hWndButtonMIDIDown=CreateWindow("button","-",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_MIDI_DOWN,hinst,NULL);

	x+=50;
	w=90;

	style=WS_CHILD|WS_VISIBLE;

	hWndLabelPoly=CreateWindow("static","ST_U",style,x,y+4,w,h,hWndForm,(HMENU)IDC_LABEL_POLY,hinst,NULL);

	x+=w;
	w=25;

	style=WS_VISIBLE|WS_CHILD|BS_DEFPUSHBUTTON|BS_OWNERDRAW|BS_NOTIFY;

	hWndButtonPolyUp=CreateWindow("button","+",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_POLY_UP,hinst,NULL);

	x+=w+5;

	hWndButtonPolyDown=CreateWindow("button","-",style,x,y,w,h,hWndForm,(HMENU)IDC_BUTTON_POLY_DOWN,hinst,NULL);

	x=10;
	y=230;
	w=GUI_WINDOW_WDT-x*2;
	h=20;

	style=WS_CHILD|WS_VISIBLE;

	hWndLabelError=CreateWindow("static","ST_U",style,x,y+4,w,h,hWndForm,(HMENU)IDC_LABEL_ERROR,hinst,NULL);

	SetUpdate(true);

	return true;
}



void GUI::close()
{
	DestroyWindow(hWndLabelV);
	DestroyWindow(hWndLabelA);
	DestroyWindow(hWndLabelP);
	DestroyWindow(hWndLabelH);
	DestroyWindow(hWndLabelD);
	DestroyWindow(hWndEditV);
	DestroyWindow(hWndEditA);
	DestroyWindow(hWndEditP);
	DestroyWindow(hWndEditH);
	DestroyWindow(hWndEditD);
	DestroyWindow(hWndButtonPulse);
	DestroyWindow(hWndButtonTri);
	DestroyWindow(hWndButtonNoise);
	DestroyWindow(hWndButtonNTSC);
	DestroyWindow(hWndButtonPAL);
	DestroyWindow(hWndButtonMultiMode);
	DestroyWindow(hWndTrackGain);
	DestroyWindow(hWndLabelMIDI);
	DestroyWindow(hWndLabelPoly);
	DestroyWindow(hWndButtonMIDIUp);
	DestroyWindow(hWndButtonMIDIDown);
	DestroyWindow(hWndButtonPolyUp);
	DestroyWindow(hWndButtonPolyDown);
	DestroyWindow(hWndLabelError);
	DestroyWindow(hWndForm);

	AEffEditor::close();
}



bool GUI::getRect(ERect **ppErect)
{
   *ppErect=&rect;

   return true;
}



void GUI::idle(void)
{
	int pgm;
	char buf[32];

	pgm=ftplug->MidiProgram[ftplug->MidiChannel];

	if(NeedUpdateEdits)
	{
		NeedUpdateEdits=false;

		SetWindowText(hWndEditV,ftplug->Program[pgm].StrVol);
		SetWindowText(hWndEditA,ftplug->Program[pgm].StrArp);
		SetWindowText(hWndEditP,ftplug->Program[pgm].StrPitch);
		SetWindowText(hWndEditH,ftplug->Program[pgm].StrHiPitch);
		SetWindowText(hWndEditD,ftplug->Program[pgm].StrDuty);

		SendMessage(hWndTrackGain,TBM_SETPOS,(WPARAM) TRUE,(LPARAM)MAX_GAIN-ftplug->OutputGain); 
	}

	if(NeedUpdate)
	{
		NeedUpdate=false;

		sprintf(buf,"MIDIch:%i",ftplug->MidiChannel+1);
		SetWindowText(hWndLabelMIDI,buf);

		sprintf(buf,"Polyphony:%i",ftplug->Polyphony);
		SetWindowText(hWndLabelPoly,buf);

		SetWindowText(hWndLabelError,CompileErrorText?CompileErrorText:"OK");

		InvalidateRect(hWndButtonMultiMode,NULL,true);
		InvalidateRect(hWndButtonNTSC ,NULL,true);
		InvalidateRect(hWndButtonPAL  ,NULL,true);
		InvalidateRect(hWndButtonPulse,NULL,true);
		InvalidateRect(hWndButtonTri  ,NULL,true);
		InvalidateRect(hWndButtonNoise,NULL,true);
		InvalidateRect(hWndLabelV,NULL,true);
		InvalidateRect(hWndLabelA,NULL,true);
		InvalidateRect(hWndLabelP,NULL,true);
		InvalidateRect(hWndLabelH,NULL,true);
		InvalidateRect(hWndLabelD,NULL,true);
	}
}



void GUI::SetUpdate(bool program)
{
	NeedUpdate=true;	
	NeedUpdateEdits=program;
}



LONG WINAPI GUI::WndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	LPDRAWITEMSTRUCT lpDIS;
	HWND phWnd;
	HDC phDC;
	GUI* gui;
	char tempstr[256];
	int len,color,pgm;
	bool compile,highlight;

	phWnd=(HWND)lParam;
	phDC =(HDC)wParam;
	
	gui =(GUI*)GetWindowLong(hWnd,GWLP_USERDATA);

	if(!gui||!gui->ftplug||!phWnd||!phDC) return DefWindowProc(hWnd,message,wParam,lParam);

	pgm=gui->ftplug->MidiProgram[gui->ftplug->MidiChannel];

	switch(message)
	{
	case WM_COMMAND:
		{
			if(HIWORD(wParam)==EN_CHANGE)	//any edit has been changed
			{
				compile=false;

				if(phWnd==gui->hWndEditV) { GetWindowText(phWnd,gui->ftplug->Program[pgm].StrVol    ,ENVELOPE_TEXT_LEN); compile=true; }
				if(phWnd==gui->hWndEditA) { GetWindowText(phWnd,gui->ftplug->Program[pgm].StrArp    ,ENVELOPE_TEXT_LEN); compile=true; }
				if(phWnd==gui->hWndEditP) { GetWindowText(phWnd,gui->ftplug->Program[pgm].StrPitch  ,ENVELOPE_TEXT_LEN); compile=true; }
				if(phWnd==gui->hWndEditH) { GetWindowText(phWnd,gui->ftplug->Program[pgm].StrHiPitch,ENVELOPE_TEXT_LEN); compile=true; }
				if(phWnd==gui->hWndEditD) { GetWindowText(phWnd,gui->ftplug->Program[pgm].StrDuty   ,ENVELOPE_TEXT_LEN); compile=true; }

				if(compile)
				{
					gui->CompileErrorText=gui->ftplug->CompileInstruments();

					gui->SetUpdate(false);
				}
			}

			if(HIWORD(wParam)==BN_CLICKED)
			{
				switch(LOWORD(wParam))
				{
				case IDC_BUTTON_MULTI:
					{	
						gui->ftplug->Program[pgm].MultiMode^=1;

						gui->SetUpdate(false);
					}
					break;

				case IDC_BUTTON_NTSC:
					{
						gui->ftplug->SysType=TVSYS_NTSC;
						gui->ftplug->SampleRatePrev=-1;

						gui->SetUpdate(false);
					}
					break;

				case IDC_BUTTON_PAL:
					{
						gui->ftplug->SysType=TVSYS_PAL;
						gui->ftplug->SampleRatePrev=-1;

						gui->SetUpdate(false);
					}
					break;

				case IDC_BUTTON_PULSE:
					{
						gui->ftplug->Program[pgm].ChType=CHTYPE_2A03_PULSE;

						gui->SetUpdate(false);
					}
					break;

				case IDC_BUTTON_TRIANGLE:
					{
						gui->ftplug->Program[pgm].ChType=CHTYPE_2A03_TRIANGLE;

						gui->SetUpdate(false);
					}
					break;

				case IDC_BUTTON_NOISE:
					{
						gui->ftplug->Program[pgm].ChType=CHTYPE_2A03_NOISE;

						gui->SetUpdate(false);
					}
					break;

				case IDC_BUTTON_MIDI_UP:
					{
						++gui->ftplug->MidiChannel;

						if(gui->ftplug->MidiChannel>=MIDI_CHANNELS_ALL) gui->ftplug->MidiChannel=0;

						gui->ftplug->UpdateGUI();
					}
					break;

				case IDC_BUTTON_MIDI_DOWN:
					{
						--gui->ftplug->MidiChannel;

						if(gui->ftplug->MidiChannel<0) gui->ftplug->MidiChannel=MIDI_CHANNELS_ALL-1;

						gui->ftplug->UpdateGUI();
					}
					break;

				case IDC_BUTTON_POLY_UP:
					{
						++gui->ftplug->Polyphony;

						if(gui->ftplug->Polyphony>SYNTH_CHANNELS) gui->ftplug->Polyphony=1;

						gui->SetUpdate(false);
					}
					break;

				case IDC_BUTTON_POLY_DOWN:
					{
						--gui->ftplug->Polyphony;

						if(gui->ftplug->Polyphony<1) gui->ftplug->Polyphony=SYNTH_CHANNELS;

						gui->SetUpdate(false);
					}
					break;
				}
			}
		}
		return 0;

	case WM_VSCROLL:
		{
			if(phWnd==gui->hWndTrackGain)
			{
				gui->ftplug->OutputGain=MAX_GAIN-SendMessage(gui->hWndTrackGain,TBM_GETPOS,0,0);
			}
		}
		return 0;

	case WM_DRAWITEM:
		{
			lpDIS=(LPDRAWITEMSTRUCT)lParam;
		
			highlight=false;

			if((lpDIS->hwndItem==gui->hWndButtonMultiMode)&&(gui->ftplug->Program[pgm].MultiMode)) highlight=true;

			if((lpDIS->hwndItem==gui->hWndButtonNTSC)&&(gui->ftplug->SysType==TVSYS_NTSC)) highlight=true;
			if((lpDIS->hwndItem==gui->hWndButtonPAL )&&(gui->ftplug->SysType==TVSYS_PAL )) highlight=true;

			if((lpDIS->hwndItem==gui->hWndButtonPulse)&&(gui->ftplug->Program[pgm].ChType==CHTYPE_2A03_PULSE   )) highlight=true;
			if((lpDIS->hwndItem==gui->hWndButtonTri  )&&(gui->ftplug->Program[pgm].ChType==CHTYPE_2A03_TRIANGLE)) highlight=true;
			if((lpDIS->hwndItem==gui->hWndButtonNoise)&&(gui->ftplug->Program[pgm].ChType==CHTYPE_2A03_NOISE   )) highlight=true;

			color=GetSysColor(highlight?COLOR_BTNSHADOW:COLOR_BTNFACE);

			SetDCBrushColor(lpDIS->hDC,color);
			SetBkColor(lpDIS->hDC,color);

			SelectObject(lpDIS->hDC,GetStockObject(DC_BRUSH));

			RoundRect(lpDIS->hDC,lpDIS->rcItem.left,lpDIS->rcItem.top,lpDIS->rcItem.right,lpDIS->rcItem.bottom,5,5);

			len=GetWindowTextLength(lpDIS->hwndItem);

			GetWindowTextA(lpDIS->hwndItem,tempstr,len+1);

			lpDIS->rcItem.top+=4;

			DrawTextA(lpDIS->hDC,tempstr,len,&lpDIS->rcItem,DT_CENTER);
		}
        return TRUE;

	case WM_CTLCOLORSTATIC:
		{
			int color=GetSysColor(COLOR_BTNFACE);

			if((phWnd==gui->hWndLabelV)&&(gui->ftplug->Program[pgm].EnvVol.error    )) color=GUI_COLOR_ERROR;
			if((phWnd==gui->hWndLabelA)&&(gui->ftplug->Program[pgm].EnvArp.error    )) color=GUI_COLOR_ERROR;
			if((phWnd==gui->hWndLabelP)&&(gui->ftplug->Program[pgm].EnvPitch.error  )) color=GUI_COLOR_ERROR;
			if((phWnd==gui->hWndLabelH)&&(gui->ftplug->Program[pgm].EnvHiPitch.error)) color=GUI_COLOR_ERROR;
			if((phWnd==gui->hWndLabelD)&&(gui->ftplug->Program[pgm].EnvDuty.error   )) color=GUI_COLOR_ERROR;

			SetBkColor(phDC,color);
		}
		return (LRESULT)GetSysColorBrush(COLOR_BTNFACE);
	}

	return DefWindowProc(hWnd,message,wParam,lParam);
}